------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: dec_by_5_iq
-- Purpose: Decimating FIR filter, with I and Q channels
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity dec_by_5_iq is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		sclkin: in STD_LOGIC;
		idatain: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		qdatain: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		romin: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		sclkout: out STD_LOGIC;
		iout: out STD_LOGIC_VECTOR (17 DOWNTO 0);
		qout: out STD_LOGIC_VECTOR (17 DOWNTO 0);
		addrout: out STD_LOGIC_VECTOR (9 DOWNTO 0)
    );
end dec_by_5_iq;

architecture rtl of dec_by_5_iq is

constant DATA_WIDTH: integer := 18;
constant DECIMATION_FACTOR: integer := 5;
constant DECIMATION_COUNT_WIDTH: integer := 4;	-- must be large enough to count to DECIMATION_FACTOR-1 
constant SAMPLE_RAM_SIZE: integer := 512;
constant SAMPLE_RAM_ADDR_WIDTH: integer := 9;	-- depends on sample ram size
constant COEFF_ROM_ADDR_WIDTH: integer := 9;  -- depends on coeff ROM size
constant OUTPUT_LSB: integer := 17;
constant ACC_WIDTH: integer := 2*DATA_WIDTH + COEFF_ROM_ADDR_WIDTH;

type SampleRAMType is array(SAMPLE_RAM_SIZE-1 downto 0) of signed ((DATA_WIDTH* 2)-1 downto 0);
 
signal StateCount: unsigned (SAMPLE_RAM_ADDR_WIDTH-1 downto 0);
signal DecimationCount: unsigned(DECIMATION_COUNT_WIDTH-1 downto 0);
signal InputAddr: unsigned(SAMPLE_RAM_ADDR_WIDTH-1 downto 0);
signal ReadAddr: unsigned(SAMPLE_RAM_ADDR_WIDTH-1 downto 0);
signal CoeffAddr: unsigned(COEFF_ROM_ADDR_WIDTH-1 downto 0);
signal SampleRAM: SampleRAMType;

signal IProd: signed ((2* DATA_WIDTH)-1 downto 0);
signal ITapData: signed (DATA_WIDTH-1 downto 0);
signal IAcc: signed (ACC_WIDTH-1 downto 0);
signal IOutreg: signed (DATA_WIDTH-1 downto 0);
signal QProd: signed ((2* DATA_WIDTH)-1 downto 0);
signal QTapData: signed (DATA_WIDTH-1 downto 0);
signal QAcc: signed (ACC_WIDTH-1 downto 0);
signal QOutreg: signed (DATA_WIDTH-1 downto 0);
signal RAMData: signed ((2* DATA_WIDTH)-1 downto 0);
signal Coeff: signed (DATA_WIDTH-1 downto 0);
signal IProdExt: signed (ACC_WIDTH-1 downto 0);
signal QProdExt: signed (ACC_WIDTH-1 downto 0);
signal IProdSign: STD_LOGIC;
signal QProdSign: STD_LOGIC;

begin --arch
  
  -- bits and pieces

iout <= STD_LOGIC_VECTOR(IOutreg);
qout <= STD_LOGIC_VECTOR(QOutreg);
addrout <= "0" & STD_LOGIC_VECTOR(CoeffAddr);
IProdSign <= IProd((DATA_WIDTH* 2)-1);
QProdSign <= QProd((DATA_WIDTH* 2)-1);
IProdExt <= IProdSign & IProdSign & IProdSign & IProdSign & IProdSign & IProdSign & IProdSign & IProdSign & IProdSign & IProd;
QProdExt <= QProdSign & QProdSign & QProdSign & QProdSign & QProdSign & QProdSign & QProdSign & QProdSign & QProdSign & QProd;


process (clk, res)
begin
	if res = '0' then
		StateCount <= (others => '0');
		DecimationCount <= (others => '0');
		InputAddr <= (others => '0');
		ReadAddr <= (others => '0');
		CoeffAddr <= (others => '0');
		IProd <= (others => '0');
		ITapData <= (others => '0');
		IAcc <= (others => '0');
		IOutreg <= (others => '0');
		QProd <= (others => '0');
		QTapData <= (others => '0');
		QAcc <= (others => '0');
		QOutreg <= (others => '0');
		--RAMData <= (others => '0');	-- initializing this stops it inferring an EBR !!
		Coeff<= (others => '0');
		sclkout <= '0';
	elsif rising_edge(clk) then
		-- handle the sample input side
		if sclkin = '1' then
			-- new input sample is ready so store it in sample RAM
			SampleRAM(to_integer(InputAddr)) <= signed(qdatain) & signed(idatain);
			InputAddr <= InputAddr + 1;
			
			if DecimationCount = (DECIMATION_FACTOR - 1) then
				DecimationCount <= (others => '0');
			else
				DecimationCount <= DecimationCount + 1;
			end if;
		end if;

		-- handle the calculation and output side
		if sclkin = '1' and DecimationCount = 0 then
			-- kick off next filter calculation, starting with newest sample
			StateCount <= (others => '0');
			ReadAddr <= InputAddr - 1;			-- don't use the current batch of samples
			CoeffAddr <= (others => '0');
		else
			-- state n-1 = set up first ram and rom addresses 
			-- state 0 = ram data available, rom addr clocked in to rom
			-- state 1 = rom data avialable, latch memory outputs
			-- state 2 = latch multiplication result
			-- state 3 = add result to accumulator

			if StateCount = 3 then
				-- multiplication results start arriving from state 3 
				IAcc <= IProdExt;
				QAcc <= QProdExt;
				-- and output previous result
				IOutreg <= IAcc(OUTPUT_LSB + DATA_WIDTH-1 downto OUTPUT_LSB);
				QOutreg <= QAcc(OUTPUT_LSB + DATA_WIDTH-1 downto OUTPUT_LSB);
				-- output sample clock pulse for previous filter output
				sclkout <= '1';
			else
				IAcc <= IAcc + IProdExt;
				QAcc <= QAcc + QProdExt;
				sclkout <= '0';
			end if;

			IProd <= Coeff * ITapData;
			QProd <= Coeff * QTapData;
			ITapData <= RAMData(DATA_WIDTH-1 downto 0);
			QTapData <= RAMData((DATA_WIDTH*2)-1 downto DATA_WIDTH);
			RAMData <= SampleRAM(to_integer(ReadAddr));
			Coeff <= signed(romin);

			-- set up addr of next most recent sample
			ReadAddr <= ReadAddr - 1;
			CoeffAddr <= CoeffAddr + 1;

			StateCount <= StateCount + 1;
		end if;

		
	end if;


end process;

end rtl;





