------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: i2c_master
-- Purpose: Master I2C interface, used to control the I2C tuner chip
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity i2c_master is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		sda_in: in STD_LOGIC;
		scl_in: in STD_LOGIC;
		start: in STD_LOGIC;
		addr_in: in STD_LOGIC_VECTOR (7 downto 0);
		reg_in: in STD_LOGIC_VECTOR (7 downto 0);
		data_in: in STD_LOGIC_VECTOR (7 downto 0);
		sda_out: out STD_LOGIC;
		scl_out: out STD_LOGIC;
		ready: out STD_LOGIC;
		data_out: out STD_LOGIC_VECTOR (7 downto 0)
    );
end i2c_master;

architecture rtl of i2c_master is

constant CLK_FREQ: integer := 96000000;
constant DELAY_COUNT_WIDTH: integer := 16;
constant DELAY_5us: integer := 5 * CLK_FREQ / 1000000;

type I2CState_type is (
	isIdle,
	isStart,
	isClockLow,
	isDataBit,
	isClockHigh,
	isRestart,
	isStop,
	isStopDelay);

signal I2CState: I2CState_type;
signal I2CDataReg: STD_LOGIC_VECTOR(7 downto 0);
signal AddrInLatch: STD_LOGIC_VECTOR(7 downto 0);
signal RegInLatch: STD_LOGIC_VECTOR(7 downto 0);
signal DataInLatch: STD_LOGIC_VECTOR(7 downto 0);
signal DataOutReg: STD_LOGIC_VECTOR(7 downto 0);
signal BitNum: unsigned(7 downto 0);
signal ByteNum: unsigned(7 downto 0);
signal DelayCount: unsigned(DELAY_COUNT_WIDTH-1 downto 0);
signal IsRead: STD_LOGIC;
signal Reading: STD_LOGIC;
signal SDAReg: STD_LOGIC;
signal SCLReg: STD_LOGIC;
signal NextDataBit: STD_LOGIC;
signal I2CReadData: STD_LOGIC_VECTOR(7 downto 0);
signal SDAInReg: STD_LOGIC;
signal SDAInReg1: STD_LOGIC;

begin --arch

sda_out <= SDAReg;
scl_out <= SCLReg;
data_out <= DataOutReg;
ready <= '1' when I2CState = isIdle else '0';


IsRead <= AddrInLatch(0);

process (clk, res)
begin
	if res = '0' then
		I2CState <= isIdle;
		I2CDataReg <= (others => '0');
		AddrInLatch <= (others => '0');
		RegInLatch <= (others => '0');
		DataInLatch <= (others => '0');
		DataOutReg <= (others => '0');
		BitNum <= (others => '0');
		ByteNum <= (others => '0');
		DelayCount <= (others => '0');
		SDAReg <= '1';
		SCLReg <= '1';
		NextDataBit <= '0';
		Reading <= '0';
	elsif rising_edge(clk) then
		if DelayCount /= 0 then
			-- We're currently in a delay period
			DelayCount <= DelayCount -1;
		else
			case I2CState is
			
			when isIdle =>
				SDAReg <= '1';
				SCLReg <= '1';
				if start = '1' then
					-- kick off a new operation
					AddrInLatch <= addr_in;
					RegInLatch <= reg_in;
					DataInLatch <= data_in;
					I2CState <= isStart;
				end if;
			
			when isStart =>
				SDAReg <= '0';		-- start condition
				DelayCount <= to_unsigned(DELAY_5us, DELAY_COUNT_WIDTH);
					-- set up to output a write command as first byte
				I2CDataReg <= AddrInLatch(7 downto 1) & '0';
				ByteNum <= (others => '0');
				BitNum <= (others => '0');
				Reading <= '0';
				I2CState <= isClockLow;
							
			when isRestart =>
				SDAReg <= '0';		-- output a second start condition
				DelayCount <= to_unsigned(DELAY_5us, DELAY_COUNT_WIDTH);
					-- now set up to output a read command (as third byte)
				I2CDataReg <= AddrInLatch(7 downto 1) & '1';
				ByteNum <= (others => '0');
				BitNum <= (others => '0');
				Reading <= '1';
				I2CState <= isClockLow;
											
			when isClockLow =>
				SCLReg <= '0';
				-- handle the input of data
				if BitNum = 0 then
					-- latch the byte of data read
					DataOutReg <= I2CReadData;
				else
					-- read in the preceding bit of data
					I2CReadData <= I2CReadData(6 downto 0) & SDAInReg;
				end if;
				-- get the next data bit ready for output...
				if BitNum = 8 then
					-- it's the 9th bit (Acknowledge bit)
					NextDataBit <= '1';		-- data high for acknowledge
					BitNum <= (others => '0');
					-- now set up next byte of data
					if Reading = '1' then
						if ByteNum = 0 then
							I2CDataReg <= x"ff";	-- leave data high ready for data input
						else
							I2CDataReg <= x"00";	-- we've read all bytes so leave data low ready for stop
						end if;
					else
						-- not reading
						if ByteNum = 0 then
							-- the second byte will be register address
							I2CDataReg <= RegInLatch;
						elsif ByteNum = 1 then
							if IsRead = '1' then
								I2CDataReg <= x"ff";	-- leave data high ready for the restart
							else
								-- if writing, the third byte will be the data byte
								I2CDataReg <= DataInLatch;
							end if;
						else
							-- ByteNum = 2
							I2CDataReg <= x"00";	-- we've sent all bytes so leave data low ready for stop
						end if;
					end if;
					ByteNum <= ByteNum + 1;
				else
					-- not the 9th (acknowledge) bit
					NextDataBit <= I2CDataReg(7);	
					I2CDataReg <= I2CDataReg(6 downto 0) & '0';		-- shift the data out msb first
					BitNum <= BitNum + 1;
				end if;
				I2CState <= isDataBit;
				
			when isDataBit =>
				-- output the next data bit, immediately after clock low
				SDAReg <= NextDataBit;
				DelayCount <= to_unsigned(DELAY_5us, DELAY_COUNT_WIDTH);
				I2CState <= isClockHigh;
			
			when isClockHigh =>
				SCLReg <= '1';
				DelayCount <= to_unsigned(DELAY_5us, DELAY_COUNT_WIDTH);
				if (Reading = '1') and (ByteNum = 2) and (BitNum = 1) then 
					I2CState <= isStop;
				elsif (IsRead = '1') and (ByteNum = 2) and (BitNum = 1) then 
					I2CState <= isRestart;
				elsif (IsRead = '0') and(ByteNum = 3) and (BitNum = 1) then 
					I2CState <= isStop;
				else
					I2CState <= isClockLow;
				end if;
						
			when isStop =>
				SDAReg <= '1';	-- stop condition
				-- add a delay before returning to idle state
				DelayCount <= to_unsigned(DELAY_5us, DELAY_COUNT_WIDTH);
				I2CState <= isStopDelay;
			
			when isStopDelay =>
				I2CState <= isIdle;
			
			when others =>
			
			end case;
		end if;
		SDAInReg <= SDAInReg1;
		SDAInReg1 <= sda_in;
	end if;  -- rising_edge(clk)
end process;


end rtl;





