------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: i2cslave
-- Purpose: Slave I2C interface for communication with the RPi
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity i2cslave is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		sda_in: in STD_LOGIC;
		scl_in: in STD_LOGIC;
		wr: in STD_LOGIC;
		data_in: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		sda_out: out STD_LOGIC;
		scl_out: out STD_LOGIC;
		rd_out: out STD_LOGIC;
		wr_out: out STD_LOGIC;
		addr_out: out STD_LOGIC_VECTOR (15 DOWNTO 0);
		data_out: out STD_LOGIC_VECTOR (7 DOWNTO 0)
    );
end i2cslave;

architecture rtl of i2cslave is

type BusStateType is (bsIdle, bsStart, bsReceiving, bsSending);

signal last_wr: STD_LOGIC;
signal last_sda: STD_LOGIC;
signal last_scl: STD_LOGIC;
signal sda: STD_LOGIC;
signal scl: STD_LOGIC;
signal sda2: STD_LOGIC;
signal scl2: STD_LOGIC;
signal SDAOutReg: STD_LOGIC;
signal Addressed: STD_LOGIC;
signal AckBitIn: STD_LOGIC;
signal BusState: BusStateType;
signal BitNum: unsigned (3 downto 0);
signal ByteNum: unsigned (7 downto 0);
signal I2CDataInReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal DataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal WrOutReg: STD_LOGIC;
signal RdOutReg: STD_LOGIC;
signal ReadCommand: STD_LOGIC;
signal I2CDataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal RegAddr: unsigned (15 DOWNTO 0);
signal NewI2CDataIn: STD_LOGIC_VECTOR (7 DOWNTO 0);

begin --arch
  
sda_out <= SDAOutReg; 
scl_out <= '1';		-- can be used for clock stretching
data_out <= DataOutReg;
addr_out <= STD_LOGIC_VECTOR(RegAddr);
wr_out <= WrOutReg;
rd_out <= RdOutReg;
NewI2CDataIn <= I2CDataInReg(6 downto 0) & sda;

process (clk, res)
begin
	-- handle monitoring of i2c bus
	if res = '0' then
		sda <= '1';
		scl <= '1';
		last_sda <= '1';
		last_scl <= '1';
		BusState <= bsIdle;
		AckBitIn <= '0';
		DataOutReg <= x"00";
		I2CDataOutReg <= x"00";
		RdOutReg <= '0';
		WrOutReg <= '0';
		SDAOutReg <= '1';
		Addressed <= '0';
		ReadCommand <= '0';
		RegAddr <= (others => '0');
	elsif rising_edge(clk) then
		if (last_sda = '1') and (sda = '0') and (last_scl = '1') and (scl = '1') then
			-- start bit
			BusState <= bsStart;
		elsif (last_sda = '0') and (sda = '1') and (last_scl = '1') and (scl = '1') then
			-- stop bit
			BusState <= bsIdle;
		else 
			case BusState is
			when bsIdle =>
				-- wait for start condition
				SDAOutReg <= '1';
				RdOutReg <= '0';
				WrOutReg <= '0';
			when bsStart =>
				SDAOutReg <= '1';
				ReadCommand <= '0';
				Addressed <= '0';
				BitNum <= (others => '0');
				ByteNum <= (others => '0');
				-- wait for SCL low
				if (scl = '0') then
					BusState <= bsReceiving;
				end if;
			when bsReceiving =>
				if (last_scl = '0') and (scl = '1') then
					-- rising edge of scl - read the data bit
					if BitNum = 7 then
						if ByteNum = 0 then	
							-- byte 0 is the I2C address and read/write flag
							if NewI2CDataIn(7 downto 1) = "1000100" then
								-- recognised our address
								Addressed <= '1';
								if NewI2CDataIn(0) = '1' then
									-- it's a read operation so fetch the first data byte to send
									-- register address will have been set by previous write op
									ReadCommand <= '1';
									RdOutReg <= '1';
								end if;
							end if;
						elsif Addressed = '1' then
							if ByteNum = 1 then
								-- Byte 1 is the register address
								RegAddr <= x"00" & unsigned(NewI2CDataIn); 
							else
								if (ByteNum >= 3) then
									RegAddr <= RegAddr + 1;
								end if;
								DataOutReg <= NewI2CDataIn;
								WrOutReg <= '1';
							end if;
						end if;
					end if;
					if BitNum <= 7 then 
						-- shift in the new bit
						I2CDataInReg <= NewI2CDataIn;
					end if;
				elsif (last_scl = '1') and (scl = '0') then
					-- falling edge of scl - need to ack ?
					if BitNum = 7 then 
						SDAOutReg <= not Addressed; -- acknowledge if we're being addressed
						BitNum <= BitNum + 1;
					elsif BitNum = 8 then
						-- got a complete byte
						if ReadCommand = '1' then
							-- it's a read command - so switch to data sending mode now
							BusState <= bsSending;
							-- set up first bit of data
							I2CDataOutReg <= data_in;
							SDAOutReg <= data_in(7);
						else
							SDAOutReg <= '1';
						end if;
						-- get ready for next byte transfer
						BitNum <= (others => '0');
						ByteNum <= ByteNum + 1;
					else
						SDAOutReg <= '1';
						BitNum <= BitNum + 1;
					end if;
				else
					-- clear read and write signals after 1 clock cycle
					WrOutReg <= '0';			
					RdOutReg <= '0';			
				end if;
				
			when bsSending =>
				if (last_scl = '0') and (scl = '1') then
					-- rising edge of scl - read the ack bit if last bit
					if BitNum = 8 then 
						AckBitIn <= sda;
						-- fetch next data byte to send
						RegAddr <= RegAddr + 1;
						RdOutReg <= '1';
						--BusState <= bsIdle;						
					end if;
					
				elsif (last_scl = '1') and (scl = '0') then
					-- falling edge of scl - need to output next bit
					if BitNum = 7 then
						SDAOutReg <= '1';  -- float the data line for acknowedge bit from master
						BitNum <= BitNum + 1;
					elsif BitNum = 8 then
						-- get ready for next byte transfer
						BitNum <= (others => '0');
						ByteNum <= ByteNum + 1;
						I2CDataOutReg <= data_in;
						-- set up first bit of data
						if AckBitIn = '0' then
							SDAOutReg <= data_in(7);
						else
							SDAOutReg <= '1';	-- no ACK was received so assume end of transfer
							BusState <= bsIdle;
						end if;
					else
						SDAOutReg <= I2CDataOutReg(6);
						I2CDataOutReg <= I2CDataOutReg(6 downto 0) & I2CDataOutReg(7);
						BitNum <= BitNum + 1;
					end if;
				else
					-- clear read and write signals after 1 clock cycle
					WrOutReg <= '0';			
					RdOutReg <= '0';			
				end if;
			end case;
		end if;

		last_sda <= sda;
		last_scl <= scl;
		sda <= sda2;	-- sync for clock domain transfer
		scl <= scl2;
		sda2 <= sda_in;
		scl2 <= scl_in;
	end if; -- rising_edge(clk)

end process;

end rtl;





