------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: io_ctrl
-- Purpose: controls the mode of operation of the physical I/O pins
-- and handles read/write commands from the SPI, I2C or serial interfaces
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity io_ctrl is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		pin_inp: in STD_LOGIC_VECTOR(47 downto 0);
		rd_in: in STD_LOGIC;
		wr_in: in STD_LOGIC;
		addr_in: in STD_LOGIC_VECTOR (15 DOWNTO 0);
		data_in: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		int_data: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		audio_clk: in STD_LOGIC;
		audio_l: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		audio_r: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		iq_clk: in STD_LOGIC;
		i_in: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		q_in: in STD_LOGIC_VECTOR (17 DOWNTO 0);
		pin_mode: out STD_LOGIC_VECTOR(47 downto 0);
		pin_dir: out STD_LOGIC_VECTOR(47 downto 0);
		pin_outp: out STD_LOGIC_VECTOR(47 downto 0);
		blink_out: out STD_LOGIC;
		int_wr_out: out STD_LOGIC;
		int_addr_out: out STD_LOGIC_VECTOR (15 DOWNTO 0);
		int_data_out: out STD_LOGIC_VECTOR (7 DOWNTO 0);
		data_out: out STD_LOGIC_VECTOR (7 DOWNTO 0);
		ctrl_out: out STD_LOGIC_VECTOR (7 DOWNTO 0);
		audio_ready: out STD_LOGIC
    );
end io_ctrl;

architecture rtl of io_ctrl is

constant VERS_HI: integer := 1;
constant VERS_LO: integer := 3;
constant CLOCK_FREQ: integer := 96000000;
constant AUDIO_CLOCK_DIV: integer := (CLOCK_FREQ/48000);
constant AUDIO_BUF_DATA_WIDTH: integer := 32; 
constant AUDIO_BUF_ADDR_WIDTH: integer := 14; 
type AudioBuf_type is array ((2**AUDIO_BUF_ADDR_WIDTH)-1 downto 0) of std_logic_vector(AUDIO_BUF_DATA_WIDTH-1 downto 0);
attribute syn_ramstyle: string;

-- Declare the Audio Buffer RAM (32 bit wide dual port RAM)
signal AudioBuf: AudioBuf_type;
signal AudioBuf_spo: STD_LOGIC_VECTOR(AUDIO_BUF_DATA_WIDTH-1 downto 0);
signal AudioBuf_dpo: STD_LOGIC_VECTOR(AUDIO_BUF_DATA_WIDTH-1 downto 0);
signal AudioBuf_di: STD_LOGIC_VECTOR(AUDIO_BUF_DATA_WIDTH-1 downto 0);
signal AudioBuf_we: STD_LOGIC;
signal AudioBuf_a: STD_LOGIC_VECTOR(AUDIO_BUF_ADDR_WIDTH-1 downto 0);
signal AudioBuf_dpra: STD_LOGIC_VECTOR(AUDIO_BUF_ADDR_WIDTH-1 downto 0);
signal AudioBuf_iptr: unsigned(AUDIO_BUF_ADDR_WIDTH-1 downto 0);
signal AudioBuf_optr: unsigned(AUDIO_BUF_ADDR_WIDTH-1 downto 0);
attribute syn_ramstyle of AudioBuf: signal is "block_ram";

signal InternalWrOutReg: STD_LOGIC;
signal InternalDataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal AddrLatch: STD_LOGIC_VECTOR (15 DOWNTO 0);
signal LowAddrLatch: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal DataIn: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal LowAddr: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal DataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal PinModeReg: STD_LOGIC_VECTOR (47 DOWNTO 0);
signal PinDirReg: STD_LOGIC_VECTOR (47 DOWNTO 0);
signal PinOutReg: STD_LOGIC_VECTOR (47 DOWNTO 0);
signal PinInReg: STD_LOGIC_VECTOR (47 DOWNTO 0);
signal PinInReg1: STD_LOGIC_VECTOR (47 DOWNTO 0);
signal CtrlReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal BlinkCount:  unsigned(31 downto 0);
signal AudioClkCount: unsigned(15 downto 0);
signal UnderSampleCount: unsigned(7 downto 0);
signal UnderSampleFactor: unsigned(7 downto 0);
signal AudioMode: STD_LOGIC_VECTOR(7 downto 0);
signal AudioSampleReg: STD_LOGIC_VECTOR(AUDIO_BUF_DATA_WIDTH-1 downto 0);
signal AudioOutReg: STD_LOGIC_VECTOR(AUDIO_BUF_DATA_WIDTH-1 downto 0);
signal AudioBufCount: STD_LOGIC_VECTOR(15 downto 0);
signal AudioEnabled: STD_LOGIC;
signal AdvanceAudioOutPtr: STD_LOGIC;


begin --architecture
  
AudioEnabled <= AudioMode(7);
LowAddrLatch <= AddrLatch(7 downto 0);
	
-- handle register reads
DataOutReg <=
	int_data when AddrLatch(15 downto 8) /= x"00"
	else PinOutReg(7 downto 0) when LowAddrLatch = x"00"
	else PinOutReg(15 downto 8) when LowAddrLatch = x"01"
	else PinOutReg(23 downto 16) when LowAddrLatch = x"02"
	else PinOutReg(31 downto 24) when LowAddrLatch = x"03"
	else PinOutReg(39 downto 32) when LowAddrLatch = x"04"
	else PinOutReg(47 downto 40) when LowAddrLatch = x"05"
	else PinDirReg(7 downto 0) when LowAddrLatch = x"10"
	else PinDirReg(15 downto 8) when LowAddrLatch = x"11"
	else PinDirReg(23 downto 16) when LowAddrLatch = x"12"
	else PinDirReg(31 downto 24) when LowAddrLatch = x"13"
	else PinDirReg(39 downto 32) when LowAddrLatch = x"14"
	else PinDirReg(47 downto 40) when LowAddrLatch = x"15"
	else PinModeReg(7 downto 0) when LowAddrLatch = x"20"
	else PinModeReg(15 downto 8) when LowAddrLatch = x"21"
	else PinModeReg(23 downto 16) when LowAddrLatch = x"22"
	else PinModeReg(31 downto 24) when LowAddrLatch = x"23"
	else PinModeReg(39 downto 32) when LowAddrLatch = x"24"
	else PinModeReg(47 downto 40) when LowAddrLatch = x"25"
	else PinInReg(7 downto 0) when LowAddrLatch = x"30"
	else PinInReg(15 downto 8) when LowAddrLatch = x"31"
	else PinInReg(23 downto 16) when LowAddrLatch = x"32"
	else PinInReg(31 downto 24) when LowAddrLatch = x"33"
	else PinInReg(39 downto 32) when LowAddrLatch = x"34"
	else PinInReg(47 downto 40) when LowAddrLatch = x"35"
	else int_data when LowAddrLatch = x"40"
	else AudioOutReg(23 downto 16) when LowAddrLatch = x"50"		-- left channel LSB
	else AudioOutReg(31 downto 24) when LowAddrLatch = x"51"		-- left channel MSB
	else AudioOutReg(7 downto 0) when LowAddrLatch = x"52"		-- right channel LSB
	else AudioOutReg(15 downto 8) when LowAddrLatch = x"53"		-- right channel MSB
	else AudioBufCount(7 downto 0) when LowAddrLatch = x"54"
	else AudioBufCount(15 downto 8) when LowAddrLatch = x"55"
	else STD_LOGIC_VECTOR(UnderSampleFactor) when LowAddrLatch = x"56"
	else AudioMode when LowAddrLatch = x"57"
	else STD_LOGIC_VECTOR(to_unsigned(VERS_HI, 8)) when LowAddrLatch = x"60"
	else STD_LOGIC_VECTOR(to_unsigned(VERS_LO, 8)) when LowAddrLatch = x"61"
	else int_data when LowAddrLatch(7 downto 4) = x"f"
	else x"00";

int_wr_out <= InternalWrOutReg;
int_addr_out <= AddrLatch;
int_data_out <= InternalDataOutReg;
pin_mode <= PinModeReg;
pin_dir <= PinDirReg;
pin_outp <= PinOutReg;
data_out <= DataOutReg;
ctrl_out <= CtrlReg;
DataIn <= data_in;
LowAddr <= addr_in(7 downto 0);


process (clk, res)
begin
	-- do blinking for LED
	if res = '0' then
		BlinkCount <= (others => '0');
		blink_out <= '0';
	elsif rising_edge(clk) then
		if BlinkCount >= (CLOCK_FREQ/2-1) then
			blink_out <= '1';
		else
			blink_out <= '0';
		end if;
		if BlinkCount >= (CLOCK_FREQ-1) then
			BlinkCount <= (others => '0');
		else
			BlinkCount <= BlinkCount + 1;
		end if;
	end if;			
end process;

--------------------------------------------------------------------------
-- Audio buffer RAM - dual port RAM with synchronous read
--------------------------------------------------------------------------
process(clk)  
begin   
	if (clk'event and clk = '1') then
		if AudioBuf_we = '1' then
			AudioBuf(to_integer(unsigned(AudioBuf_a))) <= AudioBuf_di;
		end if;
		AudioBuf_spo <= AudioBuf(to_integer(unsigned(AudioBuf_a)));
		AudioBuf_dpo <= AudioBuf(to_integer(unsigned(AudioBuf_dpra)));
	end if;
end process;
AudioBuf_dpra <= STD_LOGIC_VECTOR(AudioBuf_optr);

--------------------------------------------------------------------------
-- Audio capture
--------------------------------------------------------------------------

process (clk, res)
begin
	
		if res = '0' then
			AudioClkCount <= (others => '0');
			UnderSampleCount <= (others => '0');			
			AudioBuf_iptr <=(others => '0');			
			AudioBuf_di <=(others => '0');			
			AudioBuf_a <=(others => '0');			
			AudioBuf_we <= '0';
			AudioSampleReg <= (others => '0');
		elsif rising_edge(clk) then
			AudioSampleReg <= audio_l(17 downto 2) & audio_r(17 downto 2);
			if AudioClkCount = 0 then
				AudioClkCount <= to_unsigned (AUDIO_CLOCK_DIV - 1, 16);
				if UnderSampleCount = 0 then
					-- this is a sample we will use
					UnderSampleCount <= UnderSampleFactor - 1;
						-- write sample word to RAM buffer 
					if AudioEnabled = '1' then
						AudioBuf_we <= '1';
						AudioBuf_a <= STD_LOGIC_VECTOR(AudioBuf_iptr);
						AudioBuf_di <= AudioSampleReg;
						AudioBuf_iptr <= AudioBuf_iptr + 1;
					end if;
				else
					UnderSampleCount <= UnderSampleCount - 1;
				end if;
			else
				AudioBuf_we <= '0';
				AudioClkCount <= AudioClkCount - 1;	
			end if;
		end if;
end process;

--------------------------------------------------------------------------
-- Handle Register Reads and Writes
--------------------------------------------------------------------------

process (clk, res)
begin
	-- handle register writes
	if res = '0' then
			InternalWrOutReg <= '0';
			InternalDataOutReg <= (others => '0');
			AddrLatch <= (others => '0');
			PinModeReg <= (others => '0');
			PinDirReg <= (others => '0');
			PinOutReg <= (others => '0');
			PinInReg <= (others => '0');
			PinInReg1 <= (others => '0');
			CtrlReg <= x"02";	-- default Wideband FM
			UnderSampleFactor <= x"03";
			AdvanceAudioOutPtr <= '0';
	elsif rising_edge(clk) then
		if wr_in = '1' then
			-- a write to a register
			InternalWrOutReg <= '1';
			InternalDataOutReg <= data_in;
			AddrLatch <= addr_in;
			-- see if it's a write to a pin control register
			if addr_in(15 downto 8) = x"00" then
				if LowAddr = x"00" then PinOutReg(7 downto 0) <= DataIn;
				elsif LowAddr = x"01" then PinOutReg(15 downto 8) <= DataIn;
				elsif LowAddr = x"02" then PinOutReg(23 downto 16) <= DataIn;
				elsif LowAddr = x"03" then PinOutReg(31 downto 24) <= DataIn;
				elsif LowAddr = x"04" then PinOutReg(39 downto 32) <= DataIn;
				elsif LowAddr = x"05" then PinOutReg(47 downto 40) <= DataIn;
				elsif LowAddr = x"10" then PinDirReg(7 downto 0) <= DataIn;
				elsif LowAddr = x"11" then PinDirReg(15 downto 8) <= DataIn;
				elsif LowAddr = x"12" then PinDirReg(23 downto 16) <= DataIn;
				elsif LowAddr = x"13" then PinDirReg(31 downto 24) <= DataIn;
				elsif LowAddr = x"14" then PinDirReg(39 downto 32) <= DataIn;
				elsif LowAddr = x"15" then PinDirReg(47 downto 40) <= DataIn;
				elsif LowAddr = x"20" then PinModeReg(7 downto 0) <= DataIn; 			
				elsif LowAddr = x"21" then PinModeReg(15 downto 8) <= DataIn;
				elsif LowAddr = x"22" then PinModeReg(23 downto 16) <= DataIn;
				elsif LowAddr = x"23" then PinModeReg(31 downto 24) <= DataIn;
				elsif LowAddr = x"24" then PinModeReg(39 downto 32) <= DataIn;
				elsif LowAddr = x"25" then PinModeReg(47 downto 40) <= DataIn;
				elsif LowAddr = x"40" then CtrlReg <= DataIn;			
				elsif LowAddr = x"56" then UnderSampleFactor <= unsigned(DataIn);			
				elsif LowAddr = x"57" then AudioMode <= DataIn;			
				end if;
			end if;
		elsif rd_in = '1' then
			-- set up to read from register
			AddrLatch <= addr_in;
			InternalWrOutReg <= '0';
		else
			-- no read or write
			InternalWrOutReg <= '0';
		end if;
		
		if (rd_in = '1') and (addr_in = x"0050") then
			-- latch the next audio data and advance the output pointer
			AudioOutReg <= AudioBuf_dpo;
			AdvanceAudioOutPtr <= '1';
		else
			AdvanceAudioOutPtr <= '0';
		end if;
		
		PinInReg <= PinInReg1;	-- sync input data to allow for clock domain transfer
		PinInReg1 <= pin_inp;
		AudioBufCount <= "00" & STD_LOGIC_VECTOR(AudioBuf_iptr - AudioBuf_optr);
		if AdvanceAudioOutPtr = '1' then
			AudioBuf_optr <= AudioBuf_optr + 1;
		end if;
	end if; -- rising_edge(clk)

end process;
	
	

end rtl;





