------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: serial_intf
-- Purpose: Interfaces between UART and the IO controller, providing a
--          messaging protocol layer with CRC
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity serial_intf is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		data_in: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		tx_rdy: in STD_LOGIC;
		rx_rdy: in STD_LOGIC;
		rx_data: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		rd_out: out STD_LOGIC;
		wr_out: out STD_LOGIC;
		addr_out: out STD_LOGIC_VECTOR (15 DOWNTO 0);
		data_out: out STD_LOGIC_VECTOR (7 DOWNTO 0);
		tx_wr: out STD_LOGIC;
		tx_data: out STD_LOGIC_VECTOR (7 DOWNTO 0);
		rx_rd: out STD_LOGIC
    );
end serial_intf;

architecture rtl of serial_intf is

type RxStateType is (rxsIdle, rxsPacketStart, rxsData, rxsCRC1, rxsCRC2);
type SerialStateType is (ssIdle, ssSending, ssReceieveChar, ssReceiveDelay);

signal RxState: RxStateType;
signal ByteNum: unsigned (7 downto 0);

signal DataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal WrOutReg: STD_LOGIC;
signal RdOutReg: STD_LOGIC;
signal RxRdReg: STD_LOGIC;
signal TxWrReg: STD_LOGIC;
signal TxDataReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal CommandByte: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal ReadCommand: STD_LOGIC;
signal WriteCommand: STD_LOGIC;
signal RegisterNum: unsigned (15 DOWNTO 0);
signal DataLength: unsigned (7 DOWNTO 0);
signal AddrOutReg: unsigned (15 DOWNTO 0);
signal RxDelayCount: unsigned (7 DOWNTO 0);
signal ByteCount: unsigned (15 DOWNTO 0);
signal RxChar: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal DLEFlag: STD_LOGIC;
signal Send: STD_LOGIC;
signal SendDelayCount: unsigned (7 DOWNTO 0); 
signal CRCReg: STD_LOGIC_VECTOR(15 downto 0);
signal CRCXorVal: STD_LOGIC_VECTOR(15 downto 0);
signal CRCData: STD_LOGIC_VECTOR(7 downto 0);
signal RxCRCBitCount: unsigned(3 downto 0);
signal ReplyPacketType: STD_LOGIC_VECTOR(7 downto 0);
signal SendByteCount: unsigned (15 DOWNTO 0);
signal SendDLEFlag: STD_LOGIC;
signal SerialState: SerialStateType;
signal TxCRCReg: STD_LOGIC_VECTOR(15 downto 0);
signal TxCRCXorVal: STD_LOGIC_VECTOR(15 downto 0);
signal TxCRCData: STD_LOGIC_VECTOR(7 downto 0);
signal TxCRCBitCount: unsigned(3 downto 0);
signal AddByteToTxCRC: STD_LOGIC;
signal PacketTag: STD_LOGIC_VECTOR(7 downto 0);

begin --arch
  
data_out <= DataOutReg;
addr_out <= STD_LOGIC_VECTOR (AddrOutReg);
wr_out <= WrOutReg;
rd_out <= RdOutReg;
tx_wr <= TxWrReg;
rx_rd <= RxRdReg;
tx_data <= TxDataReg;
WriteCommand <= '1' when CommandByte = x"90" else '0';
ReadCommand <= '1' when CommandByte = x"91" else '0';
CRCXorVal <= x"1021" when (CRCReg(15) xor CRCData(7)) = '1' else x"0000";
TxCRCXorVal <= x"1021" when (TxCRCReg(15) xor TxCRCData(7)) = '1' else x"0000";

process (clk, res)
begin
	if res = '0' then
		SerialState <= ssIdle;
		RxState <= rxsIdle;
		ByteNum <= (others => '0');
		RegisterNum <= (others => '0');
		WrOutReg <= '0';
		DataOutReg <= (others => '0');
		Send <= '0';
		CRCReg <= (others => '0');
		CRCData <= (others => '0');
		RxCRCBitCount <= (others => '0');
		ReplyPacketType <= (others => '0');
		RxChar <= (others => '0');
		RxDelayCount <= (others => '0');
		RxRdReg <= '0';
		TxWrReg <= '0';
		TxDataReg <= (others => '0');
		RdOutReg <= '0';
		SendDelayCount <= (others => '0');
		SendByteCount <=  (others => '0');
		SendDLEFlag <= '0';
		AddByteToTxCRC <= '0';
		TxCRCReg <= (others => '0');
		TxCRCData <= (others => '0');
		TxCRCBitCount <= (others => '0');
		PacketTag <= (others => '0');
	elsif rising_edge(clk) then
		case SerialState is
		when ssIdle =>
			WrOutReg <= '0';
			if Send = '1' then
				-- initiate sending of a reply packet
				SendByteCount <=  (others => '0');
				SendDelayCount <= (others => '0');
				SendDLEFlag <= '0';
				RdOutReg <= '0';
				Send <= '0';
				SerialState <= ssSending;
			elsif rx_rdy = '1' then
				-- new char is ready
				RxChar <= rx_data;
				RxRdReg <= '1';	-- high for one clock only
				SerialState <= ssReceieveChar;
			end if;
			
		when ssSending =>
			-- we're in a reply phase
			if SendDelayCount /= 0 then
				-- handle CRC generation
				if AddByteToTxCRC = '1' then
					TxCRCData <= TxDataReg;
					TxCRCBitCount <= to_unsigned(8, 4);
				elsif TxCRCBitCount /= 0 then
					-- updating the CRC for next 8 bits of received data
					TxCRCReg <= (TxCRCReg(14 downto 0) & '0') xor TxCRCXorVal;
					TxCRCData <= TxCRCData(6 downto 0) & '0';
					TxCRCBitCount <= TxCRCBitCount - 1;
				end if;
				
				if SendDelayCount = 5 then
					TxWrReg <= '1';
				else	
					TxWrReg <= '0';
				end if;
				if SendDelayCount = 1 then
					if (SendByteCount = DataLength + 12) and (SendDLEFlag = '0') then
						-- finished sending
						SerialState <= ssIdle;
					end if;				
				end if;
				SendDelayCount <= SendDelayCount - 1;
				AddByteToTxCRC <= '0';
			else
				-- send next char
				if SendDLEFlag = '1' then
					-- double up the DLE
					TxDataReg <= x"10";
					SendDLEFlag <= '0';
					AddByteToTxCRC <= '0';
				else
					if SendByteCount = 0 then
						TxDataReg <= x"10";
						TxCRCReg <= (others => '0');
						AddByteToTxCRC <= '0';
					elsif SendByteCount = 1 then
						TxDataReg <= x"02";
					elsif SendByteCount = 2 then
						TxDataReg <= ReplyPacketType;
						AddByteToTxCRC <= '1';
					elsif SendByteCount = 3 then
						TxDataReg <= PacketTag;
						if PacketTag = x"10" then
							SendDLEFlag <= '1';
						end if;
						AddByteToTxCRC <= '1';
					elsif SendByteCount = 4 then
						TxDataReg <= STD_LOGIC_VECTOR(RegisterNum(7 downto 0));
						RdOutReg <= '1';
						if RegisterNum(7 downto 0) = x"10" then
							SendDLEFlag <= '1';
						end if;
						AddByteToTxCRC <= '1';
					elsif SendByteCount = 5 then
						TxDataReg <= STD_LOGIC_VECTOR(RegisterNum(15 downto 8));
						if RegisterNum(15 downto 8) = x"10" then
							SendDLEFlag <= '1';
						end if;
						AddByteToTxCRC <= '1';
					elsif SendByteCount = 6 then
						TxDataReg <= STD_LOGIC_VECTOR(DataLength);
						if DataLength = x"10" then
							SendDLEFlag <= '1';
						end if;
						AddByteToTxCRC <= '1';
					elsif SendByteCount = 7 then
						TxDataReg <= x"00";
						AddByteToTxCRC <= '1';
					elsif SendByteCount = DataLength + 8 then
						TxDataReg <= x"10";
					elsif SendByteCount = DataLength + 9 then
						TxDataReg <= x"03";
					elsif SendByteCount = DataLength + 10 then
						TxDataReg <= TxCRCReg(15 downto 8);
						if TxCRCReg(15 downto 8) = x"10" then
							SendDLEFlag <= '1';
						end if;
					elsif SendByteCount = DataLength + 11 then
						TxDataReg <= TxCRCReg(7 downto 0);
						if TxCRCReg(7 downto 0) = x"10" then
							SendDLEFlag <= '1';
						end if;
					else
						-- it's a data byte
						TxDataReg <= data_in;
						if data_in = x"10" then
							SendDLEFlag <= '1';
						end if;
						AddrOutReg <= AddrOutReg + 1;
						RdOutReg <= '1';
						AddByteToTxCRC <= '1';
					end if;
					SendByteCount <= SendByteCount + 1; 
				end if;
				SendDelayCount <= to_unsigned(10, 8);
			end if;
		
		when ssReceieveChar =>
			-- a new character has been received and is in RxChar
			-- now need to decode it
						--TxDataReg <= RxChar;	-- echo char
						--TxWrReg <= '1';
			RxRdReg <= '0';
			
			case RxState is
			when rxsIdle =>
				if RxChar = x"10" then
					RxState <= rxsPacketStart;
				end if;
			when rxsPacketStart =>
				if RxChar = x"02" then
					-- DLE STX  - start of packet
					DLEFlag <= '0';
					ByteNum <= (others => '0');
					CRCReg <= (others => '0');
					RxState <= rxsData;
				else
					RxState <= rxsIdle;		-- error
				end if;
				
			when rxsData =>
				if DLEFlag = '1' then
					DLEFlag <= '0';
					if RxChar = x"10" then
						RxState <= rxsData;
					elsif RxChar = x"03" then
						-- DLE ETX - end of pkt 
						RxState <= rxsCRC1;
					else
						  -- error
						DataLength <= (others => '0');
						ReplyPacketType <= x"20";
						RegisterNum <= x"0053";
						Send <= '1'; 
						RxState <= rxsIdle;
					end if;
				elsif RxChar = x"10" then
					DLEFlag <= '1';
				end if;
				if ((DLEFlag = '0') and (RxChar /= x"10")) or ((DLEFlag = '1') and (RxChar = x"10")) then
					if ByteNum = 0 then
						CommandByte <= RxChar;	-- packet type
					elsif ByteNum = 1 then
						PacketTag <= RxChar;
					elsif ByteNum = 2 then
						RegisterNum(7 downto 0) <= unsigned(RxChar);
					elsif ByteNum = 3 then
						-- high byte of Reg Num
						RegisterNum(15 downto 8) <= unsigned(RxChar);
					elsif ByteNum = 4 then
						DataLength <= unsigned(RxChar);
					elsif ByteNum = 5 then
						-- reserved for high byte of Data Length
					else
						if WriteCommand = '1' then
							DataOutReg <= RxChar;
							AddrOutReg <= RegisterNum;
							RegisterNum <= RegisterNum + 1;
							WrOutReg <= '1';
						end if;
					end if;
					CRCData <= RxChar;
					RxCRCBitCount <= to_unsigned(8, 4);
					ByteNum <= ByteNum + 1;					
				end if;

			when rxsCRC1 =>
				if (DLEFlag = '0') and (RxChar = x"10") then
					DLEFlag <= '1';
				elsif ((DLEFlag = '0') and (RxChar /= x"10")) or ((DLEFlag = '1') and (RxChar = x"10")) then
					DLEFlag <= '0';
					if RxChar = CRCReg(15 downto 8) then
						RxState <= rxsCRC2;
					else
						-- CRC error
						DataLength <= (others => '0');
						ReplyPacketType <= x"20";
						RegisterNum <= x"0041";
						Send <= '1'; 
						RxState <= rxsIdle;
					end if;
				else
					-- format error
					DLEFlag <= '0';
						DataLength <= (others => '0');
						ReplyPacketType <= x"20";
						RegisterNum <= x"0051";
						Send <= '1'; 
						RxState <= rxsIdle;
				end if;
				
			when rxsCRC2 =>
				if (DLEFlag = '0') and (RxChar = x"10") then
					DLEFlag <= '1';
				elsif ((DLEFlag = '0') and (RxChar /= x"10")) or ((DLEFlag = '1') and (RxChar = x"10")) then
					DLEFlag <= '0';
					if RxChar = CRCReg(7 downto 0) then
						-- got entire correct packet now
						if ReadCommand = '1' then
							-- get ready for sending the reply
							AddrOutReg <= RegisterNum;
							ReplyPacketType <= x"92";
							Send <= '1'; 
						end if;
						RxState <= rxsIdle; 
					else
						-- CRC error
						DataLength <= (others => '0');
						ReplyPacketType <= x"20";
						RegisterNum <= x"0042";
						Send <= '1'; 
						RxState <= rxsIdle;
					end if;
				else
					-- format error
					DLEFlag <= '0';
						DataLength <= (others => '0');
						ReplyPacketType <= x"20";
						RegisterNum <= x"0052";
						Send <= '1'; 
						RxState <= rxsIdle;
				end if;
			end case;
			SerialState <= ssReceiveDelay;

		when ssReceiveDelay =>
			-- delay at least one clock to allow next char to settle
			RxRdReg <= '0';
			TxWrReg <= '0';
			WrOutReg <= '0';
			if RxCRCBitCount /= 0 then
				-- updating the CRC for next 8 bits of received data
				CRCReg <= (CRCReg(14 downto 0) & '0') xor CRCXorVal;
				CRCData <= CRCData(6 downto 0) & '0';
				RxCRCBitCount <= RxCRCBitCount - 1;
			else
				SerialState <= ssIdle;
			end if;
			
		end case;
	end if; -- rising_edge(clk)							
end process;


end rtl;





