------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: spislave
-- Purpose: Slave SPI interface
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity spislave is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		sclk_in: in STD_LOGIC;
		mosi_in: in STD_LOGIC;
		cs0_in: in STD_LOGIC;
		cs1_in: in STD_LOGIC;
		data_in: in STD_LOGIC_VECTOR (7 DOWNTO 0);
		miso_out: out STD_LOGIC;
		rd_out: out STD_LOGIC;
		wr_out: out STD_LOGIC;
		addr_out: out STD_LOGIC_VECTOR (15 DOWNTO 0);
		data_out: out STD_LOGIC_VECTOR (7 DOWNTO 0)
    );
end spislave;

architecture rtl of spislave is

signal last_mosi: STD_LOGIC;
signal last_sclk: STD_LOGIC;
signal last_spi_cs: STD_LOGIC_VECTOR (1 DOWNTO 0);
signal mosi: STD_LOGIC;
signal sclk: STD_LOGIC;
signal spi_cs: STD_LOGIC_VECTOR (1 DOWNTO 0);
signal mosi2: STD_LOGIC;
signal sclk2: STD_LOGIC;
signal spi_cs2: STD_LOGIC_VECTOR (1 DOWNTO 0);
signal miso_out_reg: STD_LOGIC;
signal BitNum: unsigned (3 downto 0);
signal ByteNum: unsigned (15 downto 0);
signal SPIDataInReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal WrOutReg: STD_LOGIC;
signal RdOutReg: STD_LOGIC;
signal DataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal SPIDataOutReg: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal RegAddr: unsigned (15 DOWNTO 0);
signal CommandByte: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal NewSPIDataIn: STD_LOGIC_VECTOR (7 DOWNTO 0);
signal SPITransferActive: STD_LOGIC;
signal Reading: STD_LOGIC;
signal Writing: STD_LOGIC;

begin --arch
  
miso_out <= SPIDataOutReg(7); 
addr_out <= STD_LOGIC_VECTOR(RegAddr);
wr_out <= WrOutReg;
rd_out <= RdOutReg;
data_out <= DataOutReg;

NewSPIDataIn <= SPIDataInReg(6 downto 0) & mosi;
Reading <= '1' when CommandByte = x"01" else '0';
Writing <= '1' when CommandByte = x"00" else '0';

process (clk, res)
begin
	-- monitoring of SPI bus
	if res = '0' then
		mosi <= '0';
		sclk <= '0';
		mosi2 <= '0';
		sclk2 <= '0';
		last_mosi <= '0';
		last_sclk <= '0';
		spi_cs <= "11";
		spi_cs2 <= "11";
		last_spi_cs <= "11";
		DataOutReg <= x"00";
		SPIDataOutReg <= x"00";
		SPIDataOutReg <= x"00";
		miso_out_reg <= '0';
		SPITransferActive <= '0';
		WrOutReg <= '0';
	elsif rising_edge(clk) then
		if SPITransferActive = '0' then
			WrOutReg <= '0';
			if spi_cs /= "11" then
				-- start of SPI transfer
				BitNum <= (others => '0');
				ByteNum <= (others => '0');
				SPITransferActive <= '1';				
				SPIDataOutReg <= x"55";		-- first byte to send
			end if;
		else
			-- transfer is active
			if spi_cs = "11" then
				-- end of transfer
				WrOutReg <= '0';
				SPITransferActive <= '0';
			elsif (last_sclk = '0') and (sclk = '1') then
				-- rising edge of sclk - read the data bit
				SPIDataInReg <= NewSPIDataIn;
				if BitNum = 7 then 
					-- got a complete byte
					if ByteNum = 0 then 
						CommandByte <= NewSPIDataIn;
					elsif ByteNum = 1 then 
						RegAddr(15 downto 8) <= unsigned(NewSPIDataIn);
					elsif ByteNum = 2 then 
						RegAddr(7 downto 0) <= unsigned(NewSPIDataIn);
						if Reading = '1' then
							RdOutReg <= '1';
						end if;
					else
						if Writing = '1' then
							if (ByteNum >= 4) then
								RegAddr <= RegAddr + 1;
							end if;
							DataOutReg <= NewSPIDataIn;
							WrOutReg <= '1';
						end if;
						if Reading = '1' then
							if RegAddr = x"0051" then
								RegAddr <= x"0050";	-- special case for reading audio data
							else
								RegAddr <= RegAddr + 1;
							end if;
							RdOutReg <= '1';
						end if;
					end if;
					BitNum <= (others => '0');
					ByteNum <= ByteNum + 1;
				else
					-- more bits to come
					BitNum <= BitNum + 1;
				end if;
			elsif (last_sclk = '1') and (sclk = '0') then
				-- falling edge of sclk - set up next tx data bit
				if BitNum = 0 then
					-- set up next byte
					if (Reading = '1') and (ByteNum >= 3) then
						if spi_cs(0) = '0' then
							SPIDataOutReg <= data_in;
						else
							SPIDataOutReg <= std_logic_vector(unsigned(SPIDataInReg) + 1);  -- dummy read data
						end if;
					else
						SPIDataOutReg <= x"FF";
					end if;
				else
					-- shift out next bit
					SPIDataOutReg <= SPIDataOutReg(6 downto 0) & SPIDataOutReg(0);
				end if;
			else
				-- clear read and write signals after 1 clock cycle
				WrOutReg <= '0';			
				RdOutReg <= '0';			
			end if;
		end if;
		last_mosi <= mosi;
		last_sclk <= sclk;
		last_spi_cs <= spi_cs;
		
		mosi <= mosi2;	-- sync for clock domain transfer
		sclk <= sclk2;
		spi_cs <= spi_cs2;
		mosi2 <= mosi_in;	-- sync for clock domain transfer
		sclk2 <= sclk_in;
		spi_cs2 <= cs1_in & cs0_in;
	end if; -- rising_edge(clk)

end process;

end rtl;





