------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: tuner_ctrl
-- Purpose: Controller to initialize, retune and read the TDA tuner
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity tuner_ctrl is
    port (
		clk: in STD_LOGIC;
		res: in STD_LOGIC;
		cs_in: in STD_LOGIC;
		wr_in: in STD_LOGIC;
		addr_in: in STD_LOGIC_VECTOR (15 downto 0);
		data_in: in STD_LOGIC_VECTOR (7 downto 0);
		tun_ready: in STD_LOGIC;
		tun_data_in: in STD_LOGIC_VECTOR (7 downto 0);
		rssi_in: in STD_LOGIC_VECTOR (7 downto 0);
		
		data_out: out STD_LOGIC_VECTOR (7 downto 0);
		tun_start: out STD_LOGIC;
		tun_addr_out: out STD_LOGIC_VECTOR (7 downto 0);
		tun_reg_out: out STD_LOGIC_VECTOR (7 downto 0);
		tun_data_out: out STD_LOGIC_VECTOR (7 downto 0);
		ready_out: out STD_LOGIC;
		am_out: out STD_LOGIC;
		wfm_out: out STD_LOGIC;
		hf_out: out STD_LOGIC;
		mute_out: out STD_LOGIC
    );
end tuner_ctrl;

architecture rtl of tuner_ctrl is

constant CLK_FREQ: integer := 96000000;
constant DELAY_COUNT_WIDTH: integer := 16;
constant DELAY_100ms: STD_LOGIC_VECTOR(DELAY_COUNT_WIDTH-1 downto 0) := STD_LOGIC_VECTOR(to_unsigned(100, DELAY_COUNT_WIDTH));
constant DELAY_1s: STD_LOGIC_VECTOR(DELAY_COUNT_WIDTH-1 downto 0) := STD_LOGIC_VECTOR(to_unsigned(1000, DELAY_COUNT_WIDTH));
constant NO_DELAY: STD_LOGIC_VECTOR(DELAY_COUNT_WIDTH-1 downto 0) := STD_LOGIC_VECTOR(to_unsigned(1, DELAY_COUNT_WIDTH));
constant TUNER_ROM_ADDR_WIDTH: integer := 5;
constant TUNER_IF_FREQ: integer := 6100000;	-- IF Frequency in Hz
constant UP_CONVERTER_FREQ: integer := 96000;		-- in kHz
constant STATUS_UPDATE_PERIOD: integer := (CLK_FREQ/10);	-- 100 ms

subtype TunerRegType is STD_LOGIC_VECTOR(7 downto 0); 


-- NXP TDA18219 registers
constant ID_byte_1: 			TunerRegType := x"00";
constant ID_byte_2: 			TunerRegType := x"01";
constant ID_byte_3: 			TunerRegType := x"02";
constant Thermo_byte_1: 		TunerRegType := x"03";
constant Thermo_byte_2: 		TunerRegType := x"04";
constant Power_state_byte: 		TunerRegType := x"05";
constant Power_state_byte_2:    TunerRegType := x"06";
constant Input_power_level_byte:TunerRegType := x"07";
constant IRQ_status:            TunerRegType := x"08";
constant IRQ_enable:            TunerRegType := x"09";
constant IRQ_clear:  			TunerRegType := x"0a";
constant IRQ_set:              	TunerRegType := x"0b";
constant AGC1_byte_1:          	TunerRegType := x"0c";
constant AGC2_byte_1:          	TunerRegType := x"0d";
constant AGCK_byte_1:          	TunerRegType := x"0e";
constant RF_AGC_byte_1:        	TunerRegType := x"0f";
constant IR_MIXER_byte_1:      	TunerRegType := x"10";
constant AGC5_byte_1:          	TunerRegType := x"11";
constant IF_AGC_byte:          	TunerRegType := x"12";
constant IF_byte_1:            	TunerRegType := x"13";
constant Reference_byte:       	TunerRegType := x"14";
constant IF_Frequency_byte:    	TunerRegType := x"15";
constant RF_Frequency_byte_1:  	TunerRegType := x"16";
constant RF_Frequency_byte_2:  	TunerRegType := x"17";
constant RF_Frequency_byte_3:  	TunerRegType := x"18";
constant MSM_byte_1:           	TunerRegType := x"19";
constant MSM_byte_2:           	TunerRegType := x"1a";
constant PSM_byte_1:           	TunerRegType := x"1b";
constant DCC_byte_1:           	TunerRegType := x"1c";
constant FLO_Max_byte:         	TunerRegType := x"1d";
constant IR_Cal_byte_1:        	TunerRegType := x"1e";
constant IR_Cal_byte_2:        	TunerRegType := x"1f";
constant IR_Cal_byte_3:        	TunerRegType := x"20";
constant IR_Cal_byte_4:        	TunerRegType := x"21";
constant Vsync_Mgt_byte:       	TunerRegType := x"22";
constant IR_MIXER_byte_2:      	TunerRegType := x"23";
constant AGC1_byte_2:          	TunerRegType := x"24";
constant AGC5_byte_2:          	TunerRegType := x"25";
constant RF_Cal_byte_1:        	TunerRegType := x"26";
constant RF_Cal_byte_2:        	TunerRegType := x"27";
constant RF_Cal_byte_3:        	TunerRegType := x"28";
constant RF_Cal_byte_4:        	TunerRegType := x"29";
constant RF_Cal_byte_5:        	TunerRegType := x"2a";
constant RF_Cal_byte_6:        	TunerRegType := x"2b";
constant RF_Filter_byte_1:     	TunerRegType := x"2c";
constant RF_Filter_byte_2:     	TunerRegType := x"2d";
constant RF_Filter_byte_3:     	TunerRegType := x"2e";
constant RF_Band_Pass_Filter_byte:    TunerRegType := x"2f";
constant CP_Current_byte:      	TunerRegType := x"30";
constant AGC_Det_Out_byte:     	TunerRegType := x"31";
constant RF_AGC_Gain_byte_1:   	TunerRegType := x"32";
constant RF_AGC_Gain_byte_2:   	TunerRegType := x"33";
constant IF_AGC_Gain_byte:     	TunerRegType := x"34";
constant Power_byte_1:         	TunerRegType := x"35";
constant Power_byte_2:         	TunerRegType := x"36";
constant Misc_byte_1:          	TunerRegType := x"37";
constant rfcal_log_1:          	TunerRegType := x"38";

type TunerROMType is array (31 downto 0) of std_logic_vector (31 downto 0);

constant TunerROM : TunerROMType :=	(
    Power_state_byte_2 & x"0a" & DELAY_100ms,  -- power up
    Power_state_byte_2 & x"0a" & DELAY_100ms,  -- power up
    Power_byte_2 & x"0c" & NO_DELAY,   -- Post POR register update
    AGC1_byte_2 & x"09" & NO_DELAY,   --  Post POR register update
    RF_Filter_byte_3 & x"40" & NO_DELAY,   --  Post POR register update
    AGCK_byte_1 & x"ff" & NO_DELAY,   --  Post POR register update
    AGC5_byte_1 & x"4a" & NO_DELAY,   --  Post POR register update
    IRQ_clear & x"9f" & NO_DELAY,   --  make sure the IRQ status is reset
    MSM_byte_1 & x"3b" & NO_DELAY,   --  select which calibrations to perform
    MSM_byte_2 & x"01" & DELAY_1s,   --  launch it
    AGC1_byte_1 & x"09" & NO_DELAY,   -- 
    Reference_byte & x"00" & NO_DELAY,   --  no clock o/p

	-- transition 1   - IF and TV standard setup
    IF_Frequency_byte & STD_LOGIC_VECTOR(to_unsigned(TUNER_IF_FREQ/50000, 8)) & NO_DELAY,   --  IF freq as KHz/50
    IF_AGC_byte & x"02" & NO_DELAY,   --     p-p max out
    IF_byte_1 & x"02" & NO_DELAY,   --      HPF 0.4MHz, LPF=8MHz no offset, no IF notch
    IR_MIXER_byte_2 & x"00" & NO_DELAY,   --  HPF off, DC notch off
    AGC1_byte_1 & x"80" & NO_DELAY,   -- 
    AGC2_byte_1 & x"0f" & NO_DELAY,   -- 
    AGCK_byte_1 & x"32" & NO_DELAY,   -- 
    PSM_byte_1 & x"60" & NO_DELAY,   -- 
    RF_AGC_byte_1 & x"13" & NO_DELAY,   -- 
    IR_MIXER_byte_1 & x"0b" & NO_DELAY,   -- 
    AGC5_byte_1 & x"0b" & NO_DELAY,   -- 
    Thermo_byte_2 & x"01" & NO_DELAY,   --  turn on temperature sensor
	IRQ_clear & x"9F" & x"0001",		-- make sure the IRQ status is reset
	RF_Frequency_byte_1 & x"00" & x"0001", -- RF freq in KHz - hi - will be replaced with actual value
	RF_Frequency_byte_2 & x"00" & x"0001",	-- RF freq in KHz - mid - will be replaced with actual value
	RF_Frequency_byte_3 & x"00" & x"0001",	-- RF freq in KHz - lo - will be replaced with actual value
	MSM_byte_1 & x"41" & x"0001",		-- select which calibrations to perform
	MSM_byte_2 & x"01" & x"0001",		-- launch it - RF filters tuning, PLL locking
	x"FFFFFFFF",
	x"FFFFFFFF");
	
type TunerState_type is (
	tsIdle,
	tsTunerInit,
	tsROMSend,
	tsROMWaitReady,
	tsRead,
	tsReadWaitReady);
	
signal TunerState: TunerState_type;
signal TunerDataReg: STD_LOGIC_VECTOR(7 downto 0);
signal TunerAddrReg: STD_LOGIC_VECTOR(7 downto 0);
signal TunerRegReg: STD_LOGIC_VECTOR(7 downto 0);
signal TunerStartReg: STD_LOGIC;
signal TunerCtrlReg: STD_LOGIC_VECTOR(7 downto 0);
signal DelayCount: unsigned(DELAY_COUNT_WIDTH-1 downto 0);
signal TunerDelay: unsigned(DELAY_COUNT_WIDTH-1 downto 0);
signal ClockDiv: unsigned(31 downto 0);
signal ROMAddr: unsigned(TUNER_ROM_ADDR_WIDTH-1 downto 0);
signal TunerROMData: STD_LOGIC_VECTOR(31 downto 0); 
signal TunerFreq: STD_LOGIC_VECTOR(31 downto 0); 
signal TunerRFFreq: STD_LOGIC_VECTOR(31 downto 0); 
signal ConversionOffset: unsigned(31 downto 0); 
signal ConvertedFreq: unsigned(31 downto 0); 
signal last_tun_ready: STD_LOGIC;
signal last_wr: STD_LOGIC;
signal am: STD_LOGIC;
signal wfm: STD_LOGIC;
signal hf: STD_LOGIC;
signal SquelchReg: unsigned (7 downto 0);
signal MuteReg: STD_LOGIC_VECTOR (7 downto 0);
signal RSSIReg: unsigned (7 downto 0);
signal IsSquelched: STD_LOGIC;
signal IsMuted: STD_LOGIC;
signal RetuneReq: STD_LOGIC;
signal ReadReq: STD_LOGIC;
signal StatusReadCounter: unsigned(31 downto 0);
signal ReadByteCount: unsigned(1 downto 0);
signal RFGainByte: STD_LOGIC_VECTOR(7 downto 0);
signal IFGainByte: STD_LOGIC_VECTOR(7 downto 0);
signal GainSum: unsigned(7 downto 0);
signal AGCValue: unsigned(7 downto 0);

begin --arch

tun_addr_out <= TunerAddrReg;
tun_reg_out <= TunerRegReg;
tun_data_out <= TunerDataReg;
tun_start <= TunerStartReg;
ready_out <= '1' when TunerState = tsIdle else '0';
TunerROMData <= TunerROM(to_integer(ROMAddr));
ConversionOffset <= to_unsigned(UP_CONVERTER_FREQ, 32) when hf = '1' else to_unsigned(0, 32);
ConvertedFreq <= unsigned(TunerFreq) + ConversionOffset;
TunerRFFreq <= STD_LOGIC_VECTOR(ConvertedFreq - 25) when wfm = '0' 
  else STD_LOGIC_VECTOR(ConvertedFreq);
am <= TunerCtrlReg(0);
wfm <= TunerCtrlReg(1);
hf <= TunerCtrlReg(2);
am_out <= am;
wfm_out <= wfm;
hf_out <= hf;
mute_out <= IsSquelched or IsMuted;
IsMuted <= MuteReg(7);

data_out <= STD_LOGIC_VECTOR(RSSIReg) when addr_in(7 downto 0) = x"F8"
	else STD_LOGIC_VECTOR(AGCValue) when addr_in(7 downto 0) = x"F9"
	else RFGainByte when addr_in(7 downto 0) = x"FA"
	else IFGainByte when addr_in(7 downto 0) = x"FB"
	else x"00";

--------------------------------------------------------------------------
-- Handle Register Writes
--------------------------------------------------------------------------

process (clk, res)
begin
	-- handle register writes
	if res = '0' then
		TunerFreq <= STD_LOGIC_VECTOR(to_unsigned(93500, 32));	 -- 93.5 MHz
		TunerCtrlReg <= x"02";	-- WFM
		RetuneReq <= '0';
		ReadReq <= '0';
		StatusReadCounter <= (others => '0');
		SquelchReg <= (others => '0');
		MuteReg <= (others => '0');
	elsif rising_edge(clk) then
		if (wr_in = '1') and (last_wr = '0') and (addr_in(15 downto 8) = x"00") then
			case addr_in(7 downto 0) is
			when x"40" => 
				TunerCtrlReg <= data_in;
			when x"f0" => 
				TunerFreq(7 downto 0) <= data_in;
			when x"f1" => 
				TunerFreq(15 downto 8) <= data_in;
			when x"f2" => 
				TunerFreq(23 downto 16) <= data_in;
			when x"f3" => 
				TunerFreq(31 downto 24) <= data_in;
			when x"f4" => 
				-- command register
				if data_in = x"a0" then
					-- retune 
					RetuneReq <= '1';
				end if;
			when x"f5" => 
				SquelchReg <= unsigned(data_in);
			when x"f6" => 
				MuteReg <= data_in;
			when others => 
			end case;
		else
			if TunerState = tsROMWaitReady then
				RetuneReq <= '0';
			end if;
		end if;
		
		if StatusReadCounter = STATUS_UPDATE_PERIOD then
			-- it's time to initiate a read sequence
			StatusReadCounter <= (others => '0');
			ReadReq <= '1';
		else
			StatusReadCounter <= StatusReadCounter + 1;
			if TunerState = tsReadWaitReady then
				ReadReq <= '0';
			end if;
		end if;
		
	end if;
end process;

--------------------------------------------------------------------------
-- Handle Tuner
--------------------------------------------------------------------------

process (clk, res)
begin
	if res = '0' then
		TunerState <= tsTunerInit;
		ROMAddr <= (others => '0');
		DelayCount <= to_unsigned(2000, DELAY_COUNT_WIDTH); 		-- delay 2 seconds before initialisation
		TunerAddrReg <= (others => '0');
		TunerRegReg <= (others => '0');
		TunerDataReg <= (others => '0');
		ClockDiv <= (others => '0');
		TunerStartReg <= '0';
		IsSquelched <= '0';
	elsif rising_edge(clk) then
		if DelayCount /= 0 then
			-- We're currently in a delay period
			if ClockDiv >= (CLK_FREQ / 1000) then
				-- ClockDiv divides clock freq down to 1 kHz so we can count delays in ms
				ClockDiv <= (others => '0');
				DelayCount <= DelayCount -1;
			else
				ClockDiv <= ClockDiv + 1;
			end if;
		else
			case TunerState is
			
			when tsIdle =>
				if RetuneReq = '1' then
					ROMAddr <= to_unsigned(7, TUNER_ROM_ADDR_WIDTH);
					TunerState <= tsROMSend;
				elsif ReadReq = '1' then
					ReadByteCount <= (others => '0');
					TunerState <= tsRead;
				end if;
			
			when tsTunerInit =>
				ROMAddr <= (others => '1');
				TunerState <= tsROMSend;
				
			when tsROMSend =>
				if TunerROMData = x"FFFFFFFF" then
					TunerState <= tsIdle;	-- end of table
				elsif tun_ready = '1' then
					-- send next line of data from ROM
					TunerAddrReg <= x"C0";
					TunerRegReg <= TunerROMData (31 downto 24);
					if TunerROMData (31 downto 24) = RF_Frequency_byte_1 then
						TunerDataReg <= TunerRFFreq(23 downto 16);
					elsif TunerROMData (31 downto 24) = RF_Frequency_byte_2 then
						TunerDataReg <= TunerRFFreq(15 downto 8);
					elsif TunerROMData (31 downto 24) = RF_Frequency_byte_3 then
						TunerDataReg <= TunerRFFreq(7 downto 0);
					else
						TunerDataReg <= TunerROMData (23 downto 16);
					end if;
					TunerDelay <= unsigned(TunerROMData (15 downto 0));
					TunerStartReg <= '1';
					ROMAddr <= ROMAddr - 1;
					ClockDiv <= (others => '0');
					TunerState <= tsROMWaitReady;
				end if;	
				
			when tsROMWaitReady =>
				TunerStartReg <= '0';
				-- wait until tuner goes ready again
				if (last_tun_ready = '0') and (tun_ready = '1') then
					TunerState <= tsROMSend;
					ClockDiv <= (others => '0');
					DelayCount <= TunerDelay;  -- set up delay if any
				end if;
				
			when tsRead =>
				if ReadByteCount = "10" then
					TunerState <= tsIdle;	-- no more bytes to read
				elsif tun_ready = '1' then
					TunerAddrReg <= x"C1";
					TunerState <= tsReadWaitReady;
					TunerStartReg <= '1';
					case ReadByteCount is
						when "00" => TunerRegReg <= RF_AGC_Gain_byte_1;
						when "01" => TunerRegReg <= IF_AGC_Gain_byte;
						when others =>
					end case;
				end if;
				
			when tsReadWaitReady =>
				TunerStartReg <= '0';
				-- wait until tuner goes ready again
				if (last_tun_ready = '0') and (tun_ready = '1') then
					case ReadByteCount is
						when "00" => RFGainByte <= tun_data_in;
						when "01" => IFGainByte <= tun_data_in;
						when others =>
					end case;
					ReadByteCount <= ReadByteCount + 1;
					ClockDiv <= (others => '0');
					DelayCount <= to_unsigned(1, DELAY_COUNT_WIDTH); 		-- delay 1ms
					TunerState <= tsRead;
				end if;
			when others =>
				
			end case;
		end if;
		last_tun_ready <= tun_ready;
		last_wr <= wr_in;
		-- Check whether channel is squelched
		RSSIReg <= unsigned(rssi_in);
		if RSSIReg < SquelchReg then
			IsSquelched <= '1';
		else
			IsSquelched <= '0';
		end if;
		GainSum <= to_unsigned(19, 8) -
			(("0000" & unsigned(RFGainByte(3 downto 0))) + 
			("000000" & unsigned(RFGainByte(5 downto 4))) +
			("000000" & unsigned(IFGainByte(4 downto 3))) +
			("00000" & unsigned(IFGainByte(2 downto 0)))
			);
		AGCValue <= (GainSum + GainSum + GainSum);
	end if;  -- rising_edge(clk)
end process;


end rtl;





