------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: uartbuf5b
-- Purpose: Transmit side of UART, with 2K byte buffer
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity uartbuf5b is
    port (
		clk: in STD_LOGIC;	
		res: in STD_LOGIC;
      wr: in STD_LOGIC;
		data_in: in STD_LOGIC_VECTOR(7 downto 0);
	   tx_out: out STD_LOGIC;
      tx_ready: out STD_LOGIC;
	   iptr: out STD_LOGIC_VECTOR(10 downto 0);
	   optr: out STD_LOGIC_VECTOR(10 downto 0)
		);
end uartbuf5b;

architecture rtl of uartbuf5b is

type tx_statetype is (idle,start_bit,data_bit);
type ram_statetype is (ram_idle, reading_ram, writing_ram);
type buf_ram_type is array (2047 downto 0) of std_logic_vector(7 downto 0);

signal baud_clk: STD_LOGIC;
signal last_wr: STD_LOGIC;
signal baud_count: unsigned(15 downto 0);
signal tx_iptr: unsigned(10 downto 0);
signal tx_optr: unsigned(10 downto 0);
signal tx_reg: STD_LOGIC_VECTOR(7 downto 0);
signal bit_num: unsigned(2 downto 0);
signal tx_state: tx_statetype;


signal buf_ram: buf_ram_type;
--signal buf_ram_spo: STD_LOGIC_VECTOR(7 downto 0);
signal buf_ram_dpo: STD_LOGIC_VECTOR(7 downto 0);
signal buf_ram_di: STD_LOGIC_VECTOR(7 downto 0);
signal buf_ram_we: STD_LOGIC;
signal buf_ram_a: STD_LOGIC_VECTOR(10 downto 0);
signal buf_ram_dpra: STD_LOGIC_VECTOR(10 downto 0);

begin --arch

iptr <= STD_LOGIC_VECTOR(tx_iptr);
optr <= STD_LOGIC_VECTOR(tx_optr);
tx_ready <= '1' when ((tx_iptr - tx_optr) < 2000) else '0';

  
process(clk)  
begin
   -- implement buffer RAM 
	if (clk'event and clk = '1') then
		if buf_ram_we = '1' then
			buf_ram(to_integer(unsigned(buf_ram_a))) <= buf_ram_di;
		end if;
		--buf_ram_spo <= buf_ram(to_integer(unsigned(buf_ram_a)));
		buf_ram_dpo <= buf_ram(to_integer(unsigned(buf_ram_dpra)));
	end if;
end process;
buf_ram_dpra <= STD_LOGIC_VECTOR(tx_optr);


process(clk,res)  
begin
	-- implement baud rate clock 
	-- Divide 96 MHz by 833 to give 115,200 baud with 0.04% error
	if res = '0' then 
		baud_count <= x"0000";
		baud_clk <= '0';
	elsif rising_edge(clk) then
		if baud_count = 832 then	
			baud_count <= x"0000";
			baud_clk <= '1';
		else 
			baud_clk <= '0';
			baud_count <= baud_count + 1;
		end if;
	end if;
end process;

process(clk,res)  
begin
   if res = '0' then 
		tx_iptr <= "00000000000";
		last_wr <= '1';
		buf_ram_we <= '0';
		buf_ram_a <= (others => '0');			
		buf_ram_di <= (others => '0');			
	elsif rising_edge(clk) then
		if (wr = '1') and (last_wr = '0') then
			-- WR has gone high
			buf_ram_a <= STD_LOGIC_VECTOR(tx_iptr);
			buf_ram_we <= '1';
			buf_ram_di <= data_in;
			tx_iptr <= tx_iptr + 1;
		else
			buf_ram_we <= '0';			
		end if;	
		last_wr <= wr;
   end if;
end process;

process(clk,res)  
begin
   if res = '0' then 
      tx_out <= '1';
      bit_num <= "000";
      tx_state <= idle;
      tx_reg <= x"00";
		tx_optr <= "00000000000";
   elsif rising_edge(clk) then
		if baud_clk = '1' then

    	case tx_state is
			when idle =>
				-- output a stop bit
         	tx_out <= '1';
         	if tx_iptr /= tx_optr then
					tx_state <= start_bit;
				end if;

      	when start_bit =>
            tx_out <= '0';
            tx_reg <= buf_ram_dpo;
			tx_optr <= tx_optr + 1;
            bit_num <= "000";
				tx_state <= data_bit;

      	when data_bit =>
            tx_out <= tx_reg(0);
            tx_reg <= tx_reg(0) & tx_reg(7 downto 1);
				if bit_num = 7 then
					tx_state <= idle;
				else
            	bit_num <= bit_num + 1;
				end if;

   	end case;
		end if;
   end if;
end process;


end rtl;
