------------------------------------------------------------------------------
-- Trilby HAT - sample firmware
--
-- Copyright (c) 2005-2016 Kinetic Avionics Ltd
-- www.kinetic.co.uk
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
-- The above copyright notice and this permission notice shall be included
-- in all copies or substantial portions of the Software.
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
-- DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
-- OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- 
-- Module Name: uartrx2k
-- Purpose: Receive side of UART, with 2K byte buffer
-- Author: Keith Frewin
--
-- Version History:
-- dd/mm/yyyy Comments
-- 14/04/2016 Initial release
------------------------------------------------------------------------------


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;

entity uartrx2k is
    port (
		clk: in STD_LOGIC;	
		res: in STD_LOGIC;
      rd: in STD_LOGIC;
		serial_in: in STD_LOGIC;
      rx_ready: out STD_LOGIC;
		rx_data_out: out STD_LOGIC_VECTOR(7 downto 0)
		);
end uartrx2k;

architecture rtl of uartrx2k is

constant RX_BUF_SIZE: integer := 2048;
constant RX_BUF_ADDR_WIDTH: integer := 11; 

type rx_statetype is (rx_idle,rx_start,rx_data_bit, rx_write_to_buf, rx_done);
type rx_buf_ram_type is array (RX_BUF_SIZE-1 downto 0) of std_logic_vector(7 downto 0);

signal baud_clk: STD_LOGIC;
signal last_rd: STD_LOGIC;
signal serial_data_in: STD_LOGIC;
signal serial_data_in_2: STD_LOGIC;
signal last_serial_data_in: STD_LOGIC;
signal baud_count: unsigned(15 downto 0);
signal rx_iptr: unsigned(RX_BUF_ADDR_WIDTH-1 downto 0);
signal rx_optr: unsigned(RX_BUF_ADDR_WIDTH-1 downto 0);
signal rx_reg: STD_LOGIC_VECTOR(7 downto 0);
signal bit_num: unsigned(2 downto 0);
signal rx_state: rx_statetype;

signal rx_buf_ram: rx_buf_ram_type;
signal rx_buf_ram_spo: STD_LOGIC_VECTOR(7 downto 0);
signal rx_buf_ram_dpo: STD_LOGIC_VECTOR(7 downto 0);
signal rx_buf_ram_di: STD_LOGIC_VECTOR(7 downto 0);
signal rx_buf_ram_we: STD_LOGIC;
signal rx_buf_ram_a: STD_LOGIC_VECTOR(RX_BUF_ADDR_WIDTH-1 downto 0);
signal rx_buf_ram_dpra: STD_LOGIC_VECTOR(RX_BUF_ADDR_WIDTH-1 downto 0);


begin --arch

rx_data_out <= rx_buf_ram_dpo;
rx_ready <= '1' when rx_optr /= rx_iptr else '0';
  
-----------------------------------------------------------------------------
-- Buffer RAM processes
-----------------------------------------------------------------------------

process(clk)  
begin
   -- implement RX buffer RAM 
	if (clk'event and clk = '1') then
		if rx_buf_ram_we = '1' then
			rx_buf_ram(to_integer(unsigned(rx_buf_ram_a))) <= rx_buf_ram_di;
		end if;
		rx_buf_ram_spo <= rx_buf_ram(to_integer(unsigned(rx_buf_ram_a)));
		rx_buf_ram_dpo <= rx_buf_ram(to_integer(unsigned(rx_buf_ram_dpra)));
	end if;
end process;
rx_buf_ram_dpra <= STD_LOGIC_VECTOR(rx_optr);
  
-----------------------------------------------------------------------------
-- RX Baud rate generation
-----------------------------------------------------------------------------

process(clk,res)  
begin
   -- implement baud rate clock 
	-- Divide 96 MHz by 833 to give 115,200 baud with 0.04% error
   if res = '0' then 
      baud_count <= to_unsigned(832, 16);
      baud_clk <= '0';
   elsif rising_edge(clk) then
		if rx_state = rx_start then
			baud_count <= to_unsigned(1248, 16);	 -- start sampling 1.5 bits later
         baud_clk <= '0';
      elsif baud_count = 0 then
         baud_count <= to_unsigned(832, 16);
         baud_clk <= '1';
      else
         baud_count <= baud_count - 1;
         baud_clk <= '0';
      end if;
   end if;
end process;

-----------------------------------------------------------------------------
-- Register reads
-----------------------------------------------------------------------------

process(clk,res)  
begin
   if res = '0' then 
      rx_optr <= to_unsigned(0, RX_BUF_ADDR_WIDTH);
		last_rd <= '0';
   elsif rising_edge(clk) then
		if (rd = '0') and (last_rd = '1') then
			-- move onto next address on falling edge of rd
			rx_optr <= rx_optr + 1;
		end if;
		last_rd <= rd;
	end if;
end process;

-----------------------------------------------------------------------------
-- Receiver
-----------------------------------------------------------------------------

process(clk,res)  
begin
   if res = '0' then 
      bit_num <= "000";
      rx_state <= rx_idle;
      rx_reg <= x"00";
      rx_iptr <= to_unsigned(0, RX_BUF_ADDR_WIDTH);
		serial_data_in <= '0';
		serial_data_in_2 <= '0';
		last_serial_data_in <= '0';
		rx_buf_ram_we <= '0';
		rx_buf_ram_a <= (others =>'0');
		rx_buf_ram_di <= (others =>'0');
   elsif rising_edge(clk) then
    	case rx_state is
			when rx_idle =>
				rx_buf_ram_we <= '0';
				if (serial_data_in = '0') and (last_serial_data_in = '1') then
					rx_state <= rx_start;
				end if;

      	when rx_start =>
            bit_num <= "000";
				rx_state <= rx_data_bit;

      	when rx_data_bit =>
				if baud_clk = '1' then
            	rx_reg <= serial_data_in & rx_reg(7 downto 1);
					if bit_num = 7 then
						rx_state <= rx_write_to_buf;
					else
            		bit_num <= bit_num + 1;
					end if;
				end if;

      	when rx_write_to_buf =>
			rx_buf_ram_di <= rx_reg;
			rx_buf_ram_a <= STD_LOGIC_VECTOR(rx_iptr);
			rx_buf_ram_we <= '1';
			rx_state <= rx_done;
			
      	when rx_done =>
			rx_buf_ram_we <= '0';
			-- wait for centre of stop bit
			if baud_clk = '1' then
				rx_iptr <= rx_iptr + 1;
				rx_state <= rx_idle;
			end if;

   	end case;
		last_serial_data_in <= serial_data_in;
		serial_data_in <= serial_data_in_2;
		serial_data_in_2 <= serial_in;
   end if;
end process;


end rtl;
